// AUTO-GENERATED BY UpdateGenVersion.bat 86d03b01104b8d7f26627f3dfc6e87a4c9ef9b1c
const char GenVersion_Version[] = "0.4.4-dev";
const char GenVersion_Commit[] = "86d03b01104b8d7f26627f3dfc6e87a4c9ef9b1c";
